@@ -1,5 +1,5 @@
 /* data.c - An abstraction for data objects.
-   Copyright (C) 2002, 2003, 2004 g10 Code GmbH
+   Copyright (C) 2002, 2003, 2004, 2005, 2007 g10 Code GmbH
 
    This file is part of GPGME.
  
@@ -31,7 +31,8 @@
 #include "data.h"
 #include "util.h"
 #include "ops.h"
-#include "io.h"
+#include "priv-io.h"
+#include "debug.h"
 
 
 gpgme_error_t
@@ -57,8 +58,12 @@ _gpgme_data_new (gpgme_data_t *r_dh, struct _gpgme_data_cbs *cbs)
 void
 _gpgme_data_release (gpgme_data_t dh)
 {
-  if (dh)
-    free (dh);
+  if (!dh)
+    return;
+
+  if (dh->file_name)
+    free (dh->file_name);
+  free (dh);
 }
 
 
@@ -68,17 +73,22 @@ _gpgme_data_release (gpgme_data_t dh)
 ssize_t
 gpgme_data_read (gpgme_data_t dh, void *buffer, size_t size)
 {
+  ssize_t res;
+  TRACE_BEG2 (DEBUG_DATA, "gpgme_data_read", dh,
+	      "buffer=%p, size=%u", buffer, size);
+
   if (!dh)
     {
       errno = EINVAL;
-      return -1;
+      return TRACE_SYSRES (-1);
     }
   if (!dh->cbs->read)
     {
-      errno = EOPNOTSUPP;
-      return -1;
+      errno = ENOSYS;
+      return TRACE_SYSRES (-1);
     }
-  return (*dh->cbs->read) (dh, buffer, size);
+  res = (*dh->cbs->read) (dh, buffer, size);
+  return TRACE_SYSRES (res);
 }
 
 
@@ -88,17 +98,22 @@ gpgme_data_read (gpgme_data_t dh, void *buffer, size_t size)
 ssize_t
 gpgme_data_write (gpgme_data_t dh, const void *buffer, size_t size)
 {
+  ssize_t res;
+  TRACE_BEG2 (DEBUG_DATA, "gpgme_data_write", dh,
+	      "buffer=%p, size=%u", buffer, size);
+
   if (!dh)
     {
       errno = EINVAL;
-      return -1;
+      return TRACE_SYSRES (-1);
     }
   if (!dh->cbs->write)
     {
-      errno = EOPNOTSUPP;
-      return -1;
+      errno = ENOSYS;
+      return TRACE_SYSRES (-1);
     }
-  return (*dh->cbs->write) (dh, buffer, size);
+  res = (*dh->cbs->write) (dh, buffer, size);
+  return TRACE_SYSRES (res);
 }
 
 
@@ -108,15 +123,18 @@ gpgme_data_write (gpgme_data_t dh, const void *buffer, size_t size)
 off_t
 gpgme_data_seek (gpgme_data_t dh, off_t offset, int whence)
 {
+  TRACE_BEG2 (DEBUG_DATA, "gpgme_data_seek", dh,
+	      "offset=%lli, whence=%i", offset, whence);
+
   if (!dh)
     {
       errno = EINVAL;
-      return -1;
+      return TRACE_SYSRES (-1);
     }
   if (!dh->cbs->seek)
     {
-      errno = EOPNOTSUPP;
-      return -1;
+      errno = ENOSYS;
+      return TRACE_SYSRES (-1);
     }
 
   /* For relative movement, we must take into account the actual
@@ -128,7 +146,7 @@ gpgme_data_seek (gpgme_data_t dh, off_t offset, int whence)
   if (offset >= 0)
     dh->pending_len = 0;
 
-  return offset;
+  return TRACE_SYSRES (offset);
 }
 
 
@@ -136,6 +154,8 @@ gpgme_data_seek (gpgme_data_t dh, off_t offset, int whence)
 void
 gpgme_data_release (gpgme_data_t dh)
 {
+  TRACE (DEBUG_DATA, "gpgme_data_release", dh);
+
   if (!dh)
     return;
 
@@ -150,6 +170,8 @@ gpgme_data_release (gpgme_data_t dh)
 gpgme_data_encoding_t
 gpgme_data_get_encoding (gpgme_data_t dh)
 {
+  TRACE1 (DEBUG_DATA, "gpgme_data_get_encoding", dh,
+	  "dh->encoding=%i", dh ? dh->encoding : GPGME_DATA_ENCODING_NONE);
   return dh ? dh->encoding : GPGME_DATA_ENCODING_NONE;
 }
 
@@ -159,12 +181,58 @@ gpgme_data_get_encoding (gpgme_data_t dh)
 gpgme_error_t
 gpgme_data_set_encoding (gpgme_data_t dh, gpgme_data_encoding_t enc)
 {
+  TRACE_BEG1 (DEBUG_DATA, "gpgme_data_set_encoding", dh,
+	      "encoding=%i", enc);
   if (!dh)
-    return gpg_error (GPG_ERR_INV_VALUE);
+    return TRACE_ERR (gpg_error (GPG_ERR_INV_VALUE));
   if (enc < 0 || enc > GPGME_DATA_ENCODING_ARMOR)
-    return gpg_error (GPG_ERR_INV_VALUE);
+    return TRACE_ERR (gpg_error (GPG_ERR_INV_VALUE));
   dh->encoding = enc;
-  return 0;
+  return TRACE_ERR (0);
+}
+
+
+/* Set the file name associated with the data object with handle DH to
+   FILE_NAME.  */
+gpgme_error_t
+gpgme_data_set_file_name (gpgme_data_t dh, const char *file_name)
+{
+  TRACE_BEG1 (DEBUG_DATA, "gpgme_data_set_file_name", dh,
+	      "file_name=%s", file_name);
+
+  if (!dh)
+    return TRACE_ERR (gpg_error (GPG_ERR_INV_VALUE));
+
+  if (dh->file_name)
+    free (dh->file_name);
+
+  if (file_name)
+    {
+      dh->file_name = strdup (file_name);
+      if (!dh->file_name)
+	return TRACE_ERR (gpg_error_from_errno (errno));
+    }
+  else
+    dh->file_name = 0;
+
+  return TRACE_ERR (0);
+}
+
+
+/* Get the file name associated with the data object with handle DH,
+   or NULL if there is none.  */
+char *
+gpgme_data_get_file_name (gpgme_data_t dh)
+{
+  if (!dh)
+    {
+      TRACE (DEBUG_DATA, "gpgme_data_get_file_name", dh);
+      return NULL;
+    }
+
+  TRACE1 (DEBUG_DATA, "gpgme_data_get_file_name", dh,
+	  "dh->file_name=%s", dh->file_name);
+  return dh->file_name;
 }
 
 
@@ -177,6 +245,8 @@ _gpgme_data_inbound_handler (void *opaque, int fd)
   char buffer[BUFFER_SIZE];
   char *bufp = buffer;
   ssize_t buflen;
+  TRACE_BEG1 (DEBUG_CTX, "_gpgme_data_inbound_handler", dh,
+	      "fd=0x%x", fd);
 
   buflen = _gpgme_io_read (fd, buffer, BUFFER_SIZE);
   if (buflen < 0)
@@ -184,19 +254,19 @@ _gpgme_data_inbound_handler (void *opaque, int fd)
   if (buflen == 0)
     {
       _gpgme_io_close (fd);
-      return 0;
+      return TRACE_ERR (0);
     }
 
   do
     {
       ssize_t amt = gpgme_data_write (dh, bufp, buflen);
       if (amt == 0 || (amt < 0 && errno != EINTR))
-	return gpg_error_from_errno (errno);
+	return TRACE_ERR (gpg_error_from_errno (errno));
       bufp += amt;
       buflen -= amt;
     }
   while (buflen > 0);
-  return 0;
+  return TRACE_ERR (0);
 }
 
 
@@ -205,23 +275,25 @@ _gpgme_data_outbound_handler (void *opaque, int fd)
 {
   gpgme_data_t dh = (gpgme_data_t) opaque;
   ssize_t nwritten;
+  TRACE_BEG1 (DEBUG_CTX, "_gpgme_data_outbound_handler", dh,
+	      "fd=0x%x", fd);
 
   if (!dh->pending_len)
     {
       ssize_t amt = gpgme_data_read (dh, dh->pending, BUFFER_SIZE);
       if (amt < 0)
-	return gpg_error_from_errno (errno);
+	return TRACE_ERR (gpg_error_from_errno (errno));
       if (amt == 0)
 	{
 	  _gpgme_io_close (fd);
-	  return 0;
+	  return TRACE_ERR (0);
 	}
       dh->pending_len = amt;
     }
 
   nwritten = _gpgme_io_write (fd, dh->pending, dh->pending_len);
   if (nwritten == -1 && errno == EAGAIN)
-    return 0;
+    return TRACE_ERR (0);
 
   if (nwritten == -1 && errno == EPIPE)
     {
@@ -230,14 +302,25 @@ _gpgme_data_outbound_handler (void *opaque, int fd)
 	 end is going to tell us what happened on some other channel.
 	 Silently close our end.  */
       _gpgme_io_close (fd);
-      return 0;
+      return TRACE_ERR (0);
     }
 
   if (nwritten <= 0)
-    return gpg_error_from_errno (errno);
+    return TRACE_ERR (gpg_error_from_errno (errno));
 
   if (nwritten < dh->pending_len)
     memmove (dh->pending, dh->pending + nwritten, dh->pending_len - nwritten);
   dh->pending_len -= nwritten;
-  return 0;
+  return TRACE_ERR (0);
+}
+
+
+/* Get the file descriptor associated with DH, if possible.  Otherwise
+   return -1.  */
+int
+_gpgme_data_get_fd (gpgme_data_t dh)
+{
+  if (!dh || !dh->cbs->get_fd)
+    return -1;
+  return (*dh->cbs->get_fd) (dh);
 }
