@@ -47,7 +47,12 @@ static struct engine_ops *engine_ops[] =
   {
     &_gpgme_engine_ops_gpg,		/* OpenPGP.  */
 #ifdef ENABLE_GPGSM
-    &_gpgme_engine_ops_gpgsm		/* CMS.  */
+    &_gpgme_engine_ops_gpgsm,		/* CMS.  */
+#else
+    NULL,
+#endif
+#ifdef ENABLE_GPGCONF
+    &_gpgme_engine_ops_gpgconf		/* gpg-conf.  */
 #else
     NULL
 #endif
@@ -169,7 +174,8 @@ gpgme_get_engine_info (gpgme_engine_info_t *info)
     {
       gpgme_engine_info_t *lastp = &engine_info;
       gpgme_protocol_t proto_list[] = { GPGME_PROTOCOL_OpenPGP,
-					GPGME_PROTOCOL_CMS };
+					GPGME_PROTOCOL_CMS,
+					GPGME_PROTOCOL_GPGCONF };
       unsigned int proto;
 
       for (proto = 0; proto < DIM (proto_list); proto++)
@@ -389,8 +395,7 @@ gpgme_set_engine_info (gpgme_protocol_t proto,
 
 
 gpgme_error_t
-_gpgme_engine_new (gpgme_engine_info_t info, engine_t *r_engine,
-		   const char *lc_ctype, const char *lc_messages)
+_gpgme_engine_new (gpgme_engine_info_t info, engine_t *r_engine)
 {
   engine_t engine;
 
@@ -404,9 +409,9 @@ _gpgme_engine_new (gpgme_engine_info_t info, engine_t *r_engine,
   engine->ops = engine_ops[info->protocol];
   if (engine->ops->new)
     {
-      gpgme_error_t err = (*engine->ops->new) (&engine->engine,
-					       info->file_name, info->home_dir,
-					       lc_ctype, lc_messages);
+      gpgme_error_t err;
+      err = (*engine->ops->new) (&engine->engine,
+				 info->file_name, info->home_dir);
       if (err)
 	{
 	  free (engine);
@@ -421,6 +426,19 @@ _gpgme_engine_new (gpgme_engine_info_t info, engine_t *r_engine,
 }
 
 
+gpgme_error_t
+_gpgme_engine_reset (engine_t engine)
+{
+  if (!engine)
+    return gpg_error (GPG_ERR_INV_VALUE);
+
+  if (!engine->ops->reset)
+    return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+
+  return (*engine->ops->reset) (engine->engine);
+}
+
+
 void
 _gpgme_engine_release (engine_t engine)
 {
@@ -476,6 +494,19 @@ _gpgme_engine_set_colon_line_handler (engine_t engine,
 						 fnc, fnc_value);
 }
 
+gpgme_error_t
+_gpgme_engine_set_locale (engine_t engine, int category,
+			  const char *value)
+{
+  if (!engine)
+    return gpg_error (GPG_ERR_INV_VALUE);
+
+  if (!engine->ops->set_locale)
+    return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+
+  return (*engine->ops->set_locale) (engine->engine, category, value);
+}
+
 gpgme_error_t
 _gpgme_engine_op_decrypt (engine_t engine, gpgme_data_t ciph,
 			  gpgme_data_t plain)
@@ -685,6 +716,46 @@ _gpgme_engine_op_verify (engine_t engine, gpgme_data_t sig,
 }
 
 
+gpgme_error_t
+_gpgme_engine_op_getauditlog (engine_t engine, gpgme_data_t output,
+                              unsigned int flags)
+{
+  if (!engine)
+    return gpg_error (GPG_ERR_INV_VALUE);
+
+  if (!engine->ops->getauditlog)
+    return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+
+  return (*engine->ops->getauditlog) (engine->engine, output, flags);
+}
+
+
+gpgme_error_t
+_gpgme_engine_op_conf_load (engine_t engine, gpgme_conf_comp_t *conf_p)
+{
+  if (!engine)
+    return gpg_error (GPG_ERR_INV_VALUE);
+
+  if (!engine->ops->conf_load)
+    return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+
+  return (*engine->ops->conf_load) (engine->engine, conf_p);
+}
+
+
+gpgme_error_t
+_gpgme_engine_op_conf_save (engine_t engine, gpgme_conf_comp_t conf)
+{
+  if (!engine)
+    return gpg_error (GPG_ERR_INV_VALUE);
+
+  if (!engine->ops->conf_save)
+    return gpg_error (GPG_ERR_NOT_IMPLEMENTED);
+
+  return (*engine->ops->conf_save) (engine->engine, conf);
+}
+
+
 void
 _gpgme_engine_set_io_cbs (engine_t engine, gpgme_io_cbs_t io_cbs)
 {
