@@ -37,7 +37,6 @@ struct io_select_fd_s
   int for_read;
   int for_write;
   int signaled;
-  int frozen;
   void *opaque;
 };
 
@@ -47,7 +46,8 @@ int _gpgme_io_read (int fd, void *buffer, size_t count);
 int _gpgme_io_write (int fd, const void *buffer, size_t count);
 int _gpgme_io_pipe (int filedes[2], int inherit_idx);
 int _gpgme_io_close (int fd);
-int _gpgme_io_set_close_notify (int fd, void (*handler) (int, void *),
+typedef void (*_gpgme_close_notify_handler_t) (int,void*);
+int _gpgme_io_set_close_notify (int fd, _gpgme_close_notify_handler_t handler,
 				void *value);
 int _gpgme_io_set_nonblocking (int fd);
 
@@ -64,4 +64,13 @@ int _gpgme_io_select (struct io_select_fd_s *fds, size_t nfds, int nonblock);
    line that the child process expects.  */
 int _gpgme_io_fd2str (char *buf, int buflen, int fd);
 
+/* Duplicate a file descriptor.  This is more restrictive than dup():
+   it assumes that the resulting file descriptors are essentially
+   co-equal (for example, no private offset), which is true for pipes
+   and sockets (but not files) under Unix with the standard dup()
+   function.  Basically, this function is used to reference count the
+   status output file descriptor shared between GPGME and libassuan
+   (in engine-gpgsm.c).  */
+int _gpgme_io_dup (int fd);
+
 #endif /* IO_H */
