@@ -1,6 +1,6 @@
 /* rungpg.c - Gpg Engine.
    Copyright (C) 2000 Werner Koch (dd9jn)
-   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006 g10 Code GmbH
+   Copyright (C) 2001, 2002, 2003, 2004, 2005, 2006, 2007 g10 Code GmbH
  
    This file is part of GPGME.
  
@@ -136,6 +136,9 @@ gpg_io_event (void *engine, gpgme_event_io_t type, void *type_data)
 {
   engine_gpg_t gpg = engine;
 
+  TRACE3 (DEBUG_ENGINE, "gpgme:gpg_io_event", gpg,
+          "event %p, type %d, type_data %p",
+          gpg->io_cbs.event, type, type_data);
   if (gpg->io_cbs.event)
     (*gpg->io_cbs.event) (gpg->io_cbs.event_priv, type, type_data);
 }
@@ -463,6 +466,8 @@ gpg_new (void **engine, const char *file_name, const char *home_dir)
     goto leave;
 
   rc = _gpgme_getenv ("DISPLAY", &dft_display);
+  if (rc)
+    goto leave;
   if (dft_display)
     {
       rc = add_arg (gpg, "--display");
@@ -471,27 +476,36 @@ gpg_new (void **engine, const char *file_name, const char *home_dir)
 
       free (dft_display);
     }
-  if (rc)
-    goto leave;
 
   if (isatty (1))
     {
-      if (ttyname_r (1, dft_ttyname, sizeof (dft_ttyname)))
+      int err;
+
+      err = ttyname_r (1, dft_ttyname, sizeof (dft_ttyname));
+      if (err)
 	rc = gpg_error_from_errno (errno);
       else
 	{
-	  rc = add_arg (gpg, "--ttyname");
-	  if (!rc)
-	    rc = add_arg (gpg, dft_ttyname);
-	  if (!rc)
+          if (dft_ttyname)
+            {
+              rc = add_arg (gpg, "--ttyname");
+              if (!rc)
+                rc = add_arg (gpg, dft_ttyname);
+            }
+          else
+            rc = 0;
+          if (!rc)
 	    {
 	      rc = _gpgme_getenv ("TERM", &dft_ttytype);
-	      if (!rc)
+	      if (rc)
 		goto leave;
-
-	      rc = add_arg (gpg, "--ttytype");
-	      if (!rc)
-		rc = add_arg (gpg, dft_ttytype);
+              
+              if (dft_ttytype)
+                {
+                  rc = add_arg (gpg, "--ttytype");
+                  if (!rc)
+                    rc = add_arg (gpg, dft_ttytype);
+                }
 
 	      free (dft_ttytype);
 	    }
@@ -999,7 +1013,8 @@ read_status (engine_gpg_t gpg)
 			     received and the next thing will be that
 			     the command handler does its action.  */
 			  if (nread > 1)
-			    DEBUG0 ("ERROR, unexpected data in read_status");
+			    TRACE0 (DEBUG_CTX, "gpgme:read_status", 0,
+				    "error: unexpected data");
 
 			  add_io_cb (gpg, gpg->cmd.fd, 0,
 				     command_handler, gpg,
@@ -1032,7 +1047,6 @@ read_status (engine_gpg_t gpg)
 				    gpg->fd_data_map[gpg->cmd.linked_idx].fd;
 				  fds.for_read = 1;
 				  fds.for_write = 0;
-				  fds.frozen = 0;
 				  fds.opaque = NULL;
 				  do
 				    {
@@ -1251,6 +1265,7 @@ start (engine_gpg_t gpg)
 
   /* build the fd list for the child */
   n = 0;
+  /* The status fd is never dup'ed, so do not include it in the list.  */
   if (gpg->colon.fnc)
     {
       fd_child_list[n].fd = gpg->colon.fd[1]; 
@@ -1341,7 +1356,7 @@ start (engine_gpg_t gpg)
 	}
     }
 
-  (*gpg->io_cbs.event) (gpg->io_cbs.event_priv, GPGME_EVENT_START, NULL);
+  gpg_io_event (gpg, GPGME_EVENT_START, NULL);
   
   /* fixme: check what data we can release here */
   return 0;
@@ -2105,6 +2120,9 @@ struct engine_ops _gpgme_engine_ops_gpg =
     gpg_sign,
     gpg_trustlist,
     gpg_verify,
+    NULL,		/* getauditlog */
+    NULL,		/* conf_load */
+    NULL,		/* conf_save */
     gpg_set_io_cbs,
     gpg_io_event,
     gpg_cancel
